package PureMessage::V6::Migrate;

use strict;
use warnings;

use ActiveState::Menu qw(menu);
use ActiveState::Prompt qw(prompt yes enter);
use PureMessage::Version qw($PREFIX);
use ActiveState::Unix::ProcInfo qw(proc_info);
use Net::Domain qw(hostfqdn);
use IO::Socket::INET;

use IO::Select;
use POSIX ":sys_wait_h";
use File::Path qw(mkpath);
use Term::ANSIColor qw(color);

use PureMessage::V6::Migrate::Utils qw(
    cmd_runas_pmxv5 cmd_runas_pmxv6 is_pmxv5_installed
    has_pmxv6_db has_pmxv5_db
);

$| = 1;

my $VERSION = '1.0';

our $SCRIPT_DIR = "$PREFIX/var/migrate/v6";
my $STATE_DIR = "$PREFIX/var/migrate";
my $LOG_DIR   = "$PREFIX/var/log";
my $LOG_FILE  = "$LOG_DIR/migrate_pmxv5.log";

sub new {
    my ($class, %args) = @_;

    my $option_info = {
        ConfigFile  => {
            text        => "[Recommended] Migrate &Configuration",
            name        => 'Configuration',
            log_file    => $LOG_FILE,
            state_file  => "$STATE_DIR/configfile.state",
            once_file   => "$STATE_DIR/configfile.once",
            timeout     => 300,
        },
        LogFile    => {
            text        => "[Recommended] Migrate &Log Search Files & Data",
            name        => 'Log File',
            'requires'  => ['ConfigFile'],
            log_file    => $LOG_FILE,
            state_file  => "$STATE_DIR/logfile.state",
            once_file => "$STATE_DIR/logfile.once",
            timeout     => 300,
        },
        Quarantine      => {
            text        => "[Optional] &Database & Quarantine",
            name        => 'Database & Quarantine',
            'requires'  => ['LogFile', 'ConfigFile'],
            log_file    => $LOG_FILE,
            state_file  => "$STATE_DIR/quarantine.state",
            once_file => "$STATE_DIR/quarantine.once",
            timeout     => 600,
        },
    };

    my $self = {
        pid_file        => "$PREFIX/var/run/pmx-v6-migrate.pid",
        option_info     => $option_info,
        process_timeout => 5,
        %args,
    };

    return bless $self, $class;
}

sub start {
    my ($self) = @_;

    $self->_check_requirements();

    $self->_pid($$);
    $self->_stop_pmxv6();
    $self->_check_port_25();
    my $intro = $self->_intro_txt();
    while (1) {
        $self->_load_last_state();
        my @menu;
        push(@menu, $self->_get_option('ConfigFile'));
        push(@menu, $self->_get_option('LogFile'));
        push(@menu, $self->_get_option('Quarantine'));
        push(@menu, ["----"]);
        push(@menu, ["E&xit", sub { $self->_my_exit() }]);
        eval {
            menu(
                intro => $intro,
                menu => \@menu,
                prompt => 'Enter selection:',
                disabled_selectable => 1,
                force => 1,
            );
        };
        if ($@) {
            warn "Error while processing the menu: $@";
        }
        enter(<<EOT);

  Press enter to continue...
EOT
    }
}

sub _intro_txt {
    my ($self) = @_;
    my $intro = <<EOT;
_PureMessage v6 Migration script v ${VERSION}_

  You can only migrate from the most recent version of PureMessage (5.6.1).

  This overwrites core PureMessage 6 configuration settings and data with
  existing PureMessage 5.6.1 settings and data. Included are any passwords you
  configured to access the PureMessage Manager and the Groups Web Interface.

  Sophos Professional Services modules and settings are not imported as part
  of the migration. Contact Professional Services for assistance.

  Step 3 (Database & Quarantine) may take a very long time, depending on the
  size of your quarantine and data.

  The results of the migration will be recorded in 
  $LOG_FILE

  You are advised to review this log once the migration has completed.

  To migrate files from PureMessage 5.6.1 to PureMessage 6, select one or more
  of the following, beginning with option "1":

  Note: You cannot select option "2" (Migrate Log Search Files & Data) unless 
  option "1" (Migrate Configuration) has already been selected and run
  successfully.
EOT
    return $intro;
}

sub _check_port_25 {
    my ($self) = @_;

    print "Checking port 25...\n";

    my $fqdn = hostfqdn();

    my $is_port_open = 0;
    for my $hn ($fqdn, "localhost") {
        if (IO::Socket::INET->new(PeerAddr => "$hn:25", Timeout => 5)) {
            $is_port_open = 1;
            last;
        }
    }
    if ($is_port_open) {
        die <<EOT;

TCP port 25 is currently in use, please ensure this port is available 
before proceeding with the migration process.

EOT
    }
}

sub _stop_pmxv6 {
    my ($self) = @_;

    print "Stopping PureMessage 6...\n";

    if (-e "$PREFIX/sendmail" || -e "$PREFIX/postfix") {
        my $stop_smtp = cmd_runas_pmxv6("pmx-service stop smtp");
        $self->_run_cmd($stop_smtp, 1);
    }
    my $cmd = cmd_runas_pmxv6("pmx stop");
    $self->_run_cmd($cmd, 1);
}

sub _check_requirements {
    my ($self) = @_;

    print "\n";

    # self is running
    if ($self->_is_running()) {
        die <<EOT;
Another migration process is running. Exiting...
EOT
    }

    if (! is_pmxv5_installed()) {
        die <<EOT;
Unable to find PureMessage v5 installation. Exiting...
EOT
    }

    if (! $self->_is_pmxv5_latest()) {
        die <<EOT;
The version of PureMessage 5 or Postgres installed is out of date.
Please update to PureMessage 5.6.1 and Postgres 8.3.7 before running the migration.
Exiting...
EOT
    }

    # check pmx v5 is running
    print "Checking for PureMessage 5 running processes...\n";
    if ($self->_is_pmxv5_running() ) {
        die <<EOT;
Please stop all PureMessage 5 running processes. Exiting...
EOT
    }

    # create dirs
    mkpath($STATE_DIR) if (! -d $STATE_DIR);

    $self->_detect_db();
}

sub _detect_db {
    my ($self) = @_;

    if (has_pmxv5_db() and has_pmxv6_db()) {
        $self->{has_db} = 1;
    } else {
        $self->{has_db} = 0;
    }
}

sub _is_pmxv5_latest {
    my ($self) = @_;

    my $cmd = cmd_runas_pmxv5("$SCRIPT_DIR/pmx-v5check-version");
    my $status = $self->_run_cmd($cmd, 1);
    if ($status->{killed}) {
        print "$status->{msg}\n";
        return 0;
    } else {
        if ($status->{rc} == 0) {
            return 1;
        } elsif ($status->{rc} == 1) {
            return 0;
        } else {
            print "$status->{msg}\n";
            return 0;
        }
    }
}

sub _is_pmxv5_running {
    my ($self) = @_;

    my $cmd = cmd_runas_pmxv5("$SCRIPT_DIR/pmx-v5status");
    my $status = $self->_run_cmd($cmd, 1);
    if ($status->{killed}) {
        print "$status->{msg}\n";
        return 1;
    } else {
        if ($status->{rc} == 0) {
            return 0;
        } elsif ($status->{rc} == 1) {
            return 1;
        } else {
            print "$status->{msg}\n";
            return 1;
        }
    }
}

sub _get_option {
    my ($self, $option_name) = @_;

    my $option_info = $self->{option_info};
    my $option = $option_info->{$option_name};
    die "Unknown option name $option_name: $!" unless $option;
   
    my $enable = 0;
    my $require_modules = $option->{'requires'};
    if (defined $require_modules) {
        my $hit = 0;
        for my $require_module (@$require_modules) {
            $hit++ if $self->{ONCE}{$require_module};
        }
        $enable = 1 if $hit >= scalar(@$require_modules);
    } else {
        $enable = 1;
    }

    my $done_state = $self->{STATE}{$option_name};
    if ($enable) {
        my $txt = $option->{text};
        if ($done_state) {
            my $color = ($done_state eq 'DONE') ? 'green' : 'red';
            $txt .= ' (' . color($color) . lc($done_state) . color("reset") . ')';
        }
        my $func_ref = sub {
            $self->_run_migrate($option_name) if $self->_ask_questions($option_name);
        };
        return [$txt, $func_ref ];
    } else {
        my $new_txt = $option->{text};
        $new_txt =~ s/(.*)/\($1\)/;
        return [$new_txt, sub { $self->_disabled_menu() }];
    }
}

sub _load_last_state {
    my ($self) = @_;
    
    my $option_info = $self->{option_info};
    for my $option (keys %$option_info) {
        if (-f $option_info->{$option}{state_file}) {
            open my $r_fh, "<", $option_info->{$option}{state_file} 
                || die "Cannot open $option_info->{$option}{state_file}: $!";
			my $state_str = <$r_fh>;
            close $r_fh;
            if ($state_str) {
                chomp $state_str;
                if ($state_str eq '') {
                    $self->{STATE}{$option} = 'ERROR';
                } else {
                    $self->{STATE}{$option} = $state_str;
                    $self->_option_done($option) if ($state_str eq 'DONE');
                }
            } else {
                $self->{STATE}{$option} = 'ERROR';
            }
        } else {
            $self->{STATE}{$option} = 0;
        }

        if (-f $option_info->{$option}{once_file}) {
            $self->{ONCE}{$option} = 1;
        } else {
            $self->{ONCE}{$option} = 0;
        }
    }
}

sub _my_exit {
    my ($self) = @_;
    $self->_clear_pid();
    exit 0;
}

sub _is_running {
    my ($self) = @_;
    my $pid = $self->_pid();
    if ($pid) {
        if(kill(0, $pid)) {
            return 1;
        } else {
            $self->_clear_pid();
        }
    }
    return; 
}

sub _clear_pid {
    my ($self) = @_;
    unlink $self->{pid_file};
}

sub _pid {
    my ($self, $pid) = @_;

    my $pid_file = $self->{pid_file};
    if ($pid) {
        my $info = proc_info(root_pid => $pid);
        die "Unable to find my own pid ($pid)" unless exists $info->{$pid};

        open my $w_fh, ">$pid_file" or die "Cannot open $self->{pid_file}: $!";
        print $w_fh "$pid\n";
        close $w_fh;
    } else {
        if (open my $r_fh, "<$pid_file") {
            $pid = <$r_fh>;
            chomp $pid;
        }
    }
    $pid ||= 0;
    return $pid;
}

sub _disabled_menu {
    my ($self) = @_;

    enter(<<EOT);

  Please run the required step first before choosing this option

  Press enter to continue...
EOT
}

sub _ask_questions {
    my ($self, $option_name) = @_;
 
    my @cmd = $self->_get_cmd($option_name, 'questions');
    print "\n";
    my $status = $self->_run_cmd(\@cmd);

    if ($status->{killed}) {
        print "$status->{msg}\n";
        return 0;
    } else {
        # yes
        if ($status->{rc} == 0) {
            return 1;
        } else {
            return 0;
        }
    }
}

sub _run_cmd {
    my ($self, $cmd, $no_output) = @_;

    my $pid = fork();
    die "Failed to fork new process: $!" if !defined $pid;

    if ($pid) {
        waitpid($pid, 0);
        return $self->_check_return_status($?);
    } else {
        if ($no_output) {
            open (STDOUT, ">/dev/null") or die "Cannot redirect STDOUT: $!\n";;
            open (STDERR, ">/dev/null") or die "Cannot redirect STDERR: $!\n";;
        }
        # execute migration code (child)
        if (ref($cmd) eq 'ARRAY') {
            exec @$cmd;
            die "Failed to exec @$cmd: $!";
        } else {
            exec $cmd;
            die "Failed to exec $cmd: $!";
        }
    }
}

sub _run_migrate {
    my ($self, $option_name) = @_;
    my ($pid, $read_pipe) = $self->_get_migrate_spawn($option_name);
    $self->{last_seen} = time();

    while (1) {
        my $io = IO::Select->new();
        $io->add($read_pipe);
        if ($io->can_read(1)) {
            my $buffer = '';
            my $byte_count = sysread($read_pipe, $buffer, 1024 * 2, length($buffer));
            if ($byte_count) {
                $self->{last_seen} = time();
                $self->_print_buffer($buffer);
            } elsif ($byte_count == 0) {
                # done
                my $r_status = $self->_kill_process($pid);
                my $status = $self->_check_return_status($r_status);
                if ($status->{killed} or ($status->{rc} != 0)) {
                    print "$status->{msg}";
                }
                last;
            } else {
                $self->_kill_process($pid);
                close ($read_pipe);
                die "Unexpected error: $!";
            }
        } else {
            my $expired = $self->_check_timeout($option_name);
            if ($expired) {
                $self->_kill_process($pid);
                last;
            }
        }
    }
    close ($read_pipe);
    $self->_display_warnings();
}

sub _display_warnings {
    my ($self) = @_;

    my $warning_msgs = $self->{WARNING_MSGS};
    return unless $warning_msgs;

    print "\nWARNING: ";
    print "$_" for @$warning_msgs;
    delete $self->{WARNING_MSGS};
}

sub _option_done {
    my ($self, $option_name) = @_;

    my $option_info = $self->{option_info}{$option_name};

    my $once_file = $option_info->{once_file};
    open my $w_fh, ">$once_file" or die "Cannot create file $once_file";
    close $w_fh;
}

sub _check_return_status{
    my ($self, $return_status) = @_;

    my $msg = <<EOT;
Migration script has terminated with error.
Review the log file for more information.
EOT

    my %status;
    if ($return_status & 127) {
        my $signal = ($return_status & 127);
        $status{killed} = 1;
        $status{rc}    = undef;
        $status{msg} = $msg;
        $status{signal} = $signal;
    } elsif (($return_status >> 8) == 0) {
        $status{killed} = 0;
        $status{rc} = 0;
    } else {
        my $rc = ($return_status >> 8);
        $status{killed} = 0;
        $status{rc}    = $rc;
        $status{msg} = $msg;
    }
    return \%status;
}

sub _print_buffer {
    my ($self, $buffer) = @_;

    my @lines = split("\n", $buffer);
    # log levels: DEBUG, INFO1, INFO2, WARNING, ERROR 
    my $tab_space = '  ';
    for my $line (@lines) {
        my ($level, $msg) = ($line =~ /^(?:\[(.*)\])?\s(.*)/);
        $level = 'WARNING' unless defined $level;

        next unless $msg;
        next if $level eq 'DEBUG'; #skip debug for now

        next if ($level eq 'PING');

        if ($level =~ /^INFO(\d+)/) {
            print $tab_space x $1;
        } elsif ($level =~ /^WARNING/) {
            push @{$self->{WARNING_MSGS}}, "$msg\n";
            next;
        } else {
            print "$level: ";
        }
        print "$msg\n";
    }
}

sub _check_timeout {
    my ($self, $option_name) = @_;

    my $option_info = $self->{'option_info'};
    my $option = $option_info->{$option_name};

    my $timeout = $option->{timeout};
    my $script_name = $option->{name};

    my $cur_time = time();
    # lost communication with child
    if ($self->{last_seen} + $timeout < $cur_time) {
        my $msg = "Unable to communicate with $script_name migration script.\nWould you like to stop it (yes/no)?";
        if (yes($msg, 'no')) {
            return 1;
        }
        else {
            $self->{last_seen} = time();
        }
    }
    return;
}

sub _get_cmd {
    my ($self, $option_name, $action) = @_;

    my $module = __PACKAGE__ . '::' . $option_name;

    my $log_file = $self->{option_info}->{$option_name}{log_file};
    my $state_file = $self->{option_info}->{$option_name}{state_file};
    my $logfile_str = qq#logfile => "$log_file"#;
    my $statefile_str = qq#statefile => "$state_file"#;
    my $db_str = qq#has_db => "$self->{has_db}"#;
    my @cmd = ("$^X", "-M$module", '-e', 
        "$module->new($logfile_str, $statefile_str, $db_str)->$action()");
    my $cmd_str = "$^X -M$module -e " .
        "'$module->new($logfile_str, $statefile_str, $db_str)->$action()'";
    return wantarray ? @cmd : $cmd_str;
}

sub _get_migrate_spawn {
    my ($self, $option_name) = @_;

    my $action = 'start';
    my ($read_pipe, $write_pipe);
    pipe($read_pipe, $write_pipe);

    my $pid = fork();
    die "Failed to fork new process: $!" if !defined $pid;

    if (!$pid) {
        # execute migration code (child)
        open STDOUT, ">&=", fileno($write_pipe);
        close $read_pipe;
        close $write_pipe;
        my @cmd = $self->_get_cmd($option_name, $action); 
        exec @cmd;
        die "Failed to exec @cmd: $!";
    }
    return ($pid, $read_pipe);
}

sub _kill_process {
    my ($self, $pid) = @_;
    local $SIG{TERM} = 'IGNORE';
    kill('TERM', -$$);
    my $return;
    my $kid;
    my $count = 0;
    my $timeout = $self->{process_timeout};
    while (1) {
        $kid = waitpid(-1, WNOHANG);
        if ($kid > 0) {
            $return = $? if ($kid == $pid);
        }
        elsif ($kid == 0) {
            if ($timeout < ++$count) {
                die "Cannot stop process ($pid). Exiting ...";
            }
            # wait until kid exits
            sleep 1;
        } else {
            last;
        }
    }
    return $return;
}

1;

