package PureMessage::V6::Migrate::Base;
use strict;
use warnings;
use PureMessage::Logger;
use Carp qw(croak);
use IPC::Open3;
use IO::Select;

use PureMessage::V6::Migrate::Utils qw(
    cmd_runas_pmxv5 cmd_runas_pmxv6 cmd_runas_root is_root
);

our $NOTIFY_FD;

sub new {
    my $class = shift;
    my $self = bless { @_ }, $class;
    $self->init();
    return $self;
}

sub init {
    my $self = shift; 
    croak ("'statefile' and 'logfile' are required parameters")
        unless ($self->{statefile} && $self->{logfile});

    unless (is_root()) {
        croak $self->_notify('ERROR', "Must be run as root!");
    }

    $self->{logger} = PureMessage::Logger->new(
        log_to      => $self->{logfile},
        debug_level => 3,
    );

    # Initialize once
    unless (defined $NOTIFY_FD) {
        open($NOTIFY_FD, ">&STDOUT") ||
            die("Cannot dup STDOUT: $!\n");
        select((select($NOTIFY_FD), $| = 1)[0]);
    }

    $SIG{__WARN__} = sub { $self->log_err(shift) };
    return;
}

sub redirect_fds {
    my $self = shift;

    unless ($self->{redirected}) {
        open(STDIN, ">/dev/null") ||
            die("Cannot redirect STDIN: $!\n");

        open(STDOUT, ">/dev/null") ||
            die("Cannot redirect STDOUT: $!\n");

        open(STDERR, ">/dev/null") ||
            die("Cannot redirect STDERR: $!\n");

        $self->{redirected}++;
    }

    return;
}

sub questions {
    my $self = shift;

    my $rc = eval { $self->display(@_) };
    if ($@) {
        $self->log2file("ERROR", "Failed before executing migration step: $@");
        print STDOUT <<EOT;
Failed before executing migration step. Please review migration log and try again.
EOT
    }

    $self->_exit_code(0) if $rc;
    return $self->_exit_code();
}

# If not defined in sub-class.
sub display {
    return 1;
}

sub start {
    my $self = shift;

    my $type  = (split "::", ref($self))[-1];
    $self->log2file("INFO", "Starting migration for $type");

    $self->redirect_fds();

    $self->done();
    my $rc = eval { $self->migrate(@_) };
    if ($@) {
        $self->log_err($@);
        $self->done(state => 1);
    }
    elsif (!$rc) {
        $self->done(state => 1);
    }
    else {
        $self->_exit_code(0);
        $self->done(state => 0);
    }

    return $self->_exit_code(); 
}

#
# Logging
#

sub show {
    my ($self, $msg) = @_;
    $self->log2file('INFO', $msg);
    print STDOUT "$msg\n";
    return;
}

sub log_debug  { shift->_log('DEBUG', @_) }
sub log_info1  { shift->_log('INFO1', @_) }
sub log_info2  { shift->_log('INFO2', @_) }
sub log_warn   { shift->_log('WARNING', @_) }
sub log_err    { shift->_log('ERROR',  @_) }
sub ping       { shift->_notify('PING', 'ping') }

sub _notify_fd { $NOTIFY_FD }

sub _notify {
    my ($self, $type, $msg) = @_;

    if ($type eq 'WARNING') {
        my @msgs = split(/\n/, $msg);
        print {$NOTIFY_FD} "[$type] $_\n" for @msgs;
        return;
    }

    $msg =~ s/^([^\r\n]*).*/$1/s;
    print {$NOTIFY_FD} "[$type] $msg\n";
    return;
}

sub log2file {
    my ($self, $type, $msg) = @_;
    $type = 'INFO' if $type =~ /^INFO/;

    # Strip ANSI escapes
    $msg =~ s/\x1b\x5b.+?[\x40-\x7e]//g;

    $msg =~ s/^\s*//s;
    $msg =~ s/\s*$//s;
    $self->{logger}->log($type, $msg);
    return;
}

sub _log {
    my ($self, $type, $msg) = @_;
    $self->log2file($type, $msg);
    $self->_notify($type, $msg);
    return;
}

#
# Run & Capture
#
sub run_as_pmxv5 {
    my ($self, $cmd, %opts) = @_;
    my $cmdline = cmd_runas_pmxv5($cmd, %opts);
    return $self->_run($cmdline, %opts);
}

sub run_as_pmxv6 {
    my ($self, $cmd, %opts) = @_;
    my $cmdline = cmd_runas_pmxv6($cmd, %opts);
    return $self->_run($cmdline, %opts);
}

sub run_as_root {
    my ($self, $cmd, %opts) = @_;
    my $cmdline = cmd_runas_root($cmd, %opts);
    return $self->_run($cmdline, %opts);
}

sub pipeline_run {
    my ($self, $cmd_ref, $cref, %opts) = @_;

    my @cmdline = @$cmd_ref;
    $self->log2file("INFO", "Running '@cmdline'");

    use Symbol 'gensym'; 

    my $stdout_fh;
    my $stderr_fh = gensym; 
    my $pid = open3(undef, $stdout_fh, $stderr_fh, @cmdline);
    die("Unable to run command '@cmdline': $!\n") unless $pid;

    my @error_lines;
    my @other_lines;
    $self->enumerate_fh($stdout_fh, $stderr_fh, sub {
        my ($err, $line) = @_;
        if ($err) {
            if ($line =~ s/^MIG_ERR:\t//) {
                push @error_lines, $line;
            } else {
                push @other_lines, $line;
            }
        } else {
            if ((not exists $opts{filter}) || $line =~ $opts{filter}) {
                $cref->($line);
            } 
        }
    });
    close ($stdout_fh);
    close ($stderr_fh);

    waitpid($pid, 0);
    my $exit_code = $? >> 8;

    if ($exit_code) {
        my @die_msgs;
        push @die_msgs, "Command failed ($exit_code): '@cmdline'\n";
        push @die_msgs, @error_lines;
        push @die_msgs, @other_lines unless @error_lines;
        my $die_msg = join('', @die_msgs);
        die "$die_msg\n";
    }

    return;
}

sub pipeline_as_pmxv5 {
    my ($self, $cmd, $cref, %opts) = @_;
    croak("pipeline_as_pmxv5: Missing parameters") unless $cmd && $cref;

    my @cmdline = cmd_runas_pmxv5($cmd, %opts);

    $self->pipeline_run(\@cmdline, $cref, %opts);
    return;
}

sub pipeline_as_pmxv6 {
    my ($self, $cmd, $cref, %opts) = @_;
    croak("pipeline_as_pmxv6: Missing parameters") unless $cmd && $cref;

    my @cmdline = cmd_runas_pmxv6($cmd, %opts);

    $self->pipeline_run(\@cmdline, $cref, %opts);
    return;
}

sub enumerate_fh {
    my ($self, $out_fh, $err_fh, $cref) = @_;

    return unless ($out_fh and $err_fh);

    my $timeout = 120;
    my $s = IO::Select->new();

    $s->add($out_fh);
    $s->add($err_fh);

    my $stderr = '';
    my $stdout = '';

    NEXT_SELECT: while (my @ready = $s->can_read($timeout)) {
        NEXT_FD: for my $fh (@ready) {
            my $buf;
            my $is_err = (fileno($fh) == fileno($err_fh));
            my $line = ($is_err) ? \$stderr : \$stdout;

            if ((sysread($fh, $buf, 50)) == 0) {
                $cref->($is_err, "$$line\n") if $$line;
		$s->remove($fh);
                next NEXT_FD;
            }
            
            $$line .= $buf;
            while ($$line =~ m{[\n\r]}g) {
                 my $offset  = pos($$line);
                 my $segment = substr($$line, 0, $offset-1);

                 $cref->($is_err, "$segment\n") if $segment;
                 $$line = ($offset == length($$line)) ? "" : substr($$line, $offset);
            }
        }
    }

    return;
}

sub _run {
    my ($self, $cmd, %opts) = @_;

    $self->log2file("INFO", "Running '$cmd'");
    my @output = qx($cmd 2>&1);
    my $exit   = $?;

    $self->log2file("INFO", $_) for @output;
    my $exit_code = $exit >> 8;
    if ($exit_code) {
        my %exceptions = map { $_ => 1 } ($opts{exitcodes})
                         ?  @{$opts{exitcodes}} : ();
        unless (exists $exceptions{$exit_code}) {
            my $msg = @output ? $output[0] : '<no output>';
            $! = 0; $? = $exit_code;
            die "Command failed ($exit_code): '$cmd': $msg\n";
        }
    }
    return $exit_code;
}

#
# Exit state
#
sub done {
    my $self = shift;
    my %opts = @_;

    open(my $fh, ">", $self->{statefile}) ||
        die("Can't create '$self->{statefile}': $!\n");

    if (exists $opts{state}) {
        my $state = $opts{state} ? 'error' : 'done';
        print {$fh} uc($state);
        $self->log2file("INFO", "State is '" . ucfirst($state) . "'");

        if ($state eq 'error') {
            my $detail_info = <<EOT;
This error could be caused by permissions or security problems.  Please ensure
    root's umask is set to 022 as per the documentation. If you are running
    SE Linux, check your audit log and please ensure that allow_execmod is
    permitted.
EOT

            $self->log2file("INFO", $detail_info );
        }
    }

    close($fh) || die("Can't close '$self->{statefile}': $!\n");
    return;
}

sub _exit_code {
    my $self = shift;
    if (@_) {
        $self->{EXIT_CODE} = shift;
    }
    return exists $self->{EXIT_CODE} ? $self->{EXIT_CODE} : 1;
}

sub DESTROY {
    my $self = shift;
    unless ($self->{no_exit}) {
        exit $self->_exit_code();
    }
    return;
}

1;
