package PureMessage::V6::Migrate::ConfigFile;
use strict;
use warnings;
use base 'PureMessage::V6::Migrate::Base';
use List::Util                      qw(first);
use ActiveState::Prompt             qw(yes);
use PureMessage::Version            qw($PREFIX);
use PureMessage::V6::Migrate::Utils qw(copy);
use Fcntl qw(S_IMODE);

our $MANIFEST_FILE = "$PREFIX/var/tmp/MIGRATION_CONFIG_MANIFEST.txt";
our $SCRIPT_DIR = "$PREFIX/var/migrate/v6";
our $IMAGE_DIR  = "$PREFIX/var/migrate/images";

sub display {
    my $self = shift;

    $self->show("Migrating your PureMessage 5 configuration will " .
                "overwrite your current settings."); 

    if (yes("Would you like to proceed (yes/no)?", 1)) {
        return 1;
    } 
    return 0;
}

sub migrate {
    my $self = shift;
    my %opts = @_;

    if (!$self->backup_mta_config()) {
        $self->log_err("Backing up MTA Config");
        return 0;
    }
    elsif (!$self->process_files(%opts)) {
        $self->log_err("Copying files");
        return 0;
    }
    elsif (!$self->fix_mta_config()) {
        $self->log_err("Restoring MTA Config");
        return 0;
    }
    elsif (!$self->finalize()) {
        $self->log_err("Rebuilding configs");
        return 0;
    }
    elsif (!$self->validate()) {
        $self->log_err("Validating configs");
        return 0;
    }

    return 1;
}

#####

sub backup_mta_config {
    my ($self) = @_;
    
    $self->run_as_pmxv6("$SCRIPT_DIR/pmx-v6mta-config --backup");
    return 1;
}

sub fix_mta_config {
    my ($self) = @_;

    my $mta = $self->mta_type();
    if ($mta eq 'postfix') {
        return $self->fix_postfix_config();
    } elsif ($mta eq 'sendmail') {
        return $self->fix_sendmail_config();
    }
    return 1;
}

sub fix_postfix_config {
    my ($self) = @_;

    my @postfix_confs;
    $self->pipeline_as_pmxv6(
        "$SCRIPT_DIR/pmx-v6mta-config --display",
        sub {
            chomp (my $line = shift || '');
            my $data = (split /\t/, $line, 2)[-1];
            push @postfix_confs, $data;
        },
        filter => qr/MTA_CONFIG:\t/,
    );

    die "Missing postfix config value.\n" unless (@postfix_confs);

    for my $postfix_conf (@postfix_confs) {
        $self->run_as_root("$PREFIX/postfix/sbin/postconf -e '$postfix_conf'");
    }

    return 1;
}

sub fix_sendmail_config {
    my ($self) = @_;

    my $type;
    $self->pipeline_as_pmxv6(
        "$SCRIPT_DIR/pmx-v6mta-config --display",
        sub {
            chomp (my $line = shift || '');
            $type = (split /\t/, $line, 2)[-1];
        },
        filter => qr/MTA_CONFIG:\t/,
    );

    die "Missing sendmail config value.\n" unless ($type);

    # only try to fix for ipv6
    return 1 unless ($type eq 'IPV6');

    my $sendmail_mc = "$PREFIX/sendmail/etc/mail/sendmail.mc";
    my @sendmail_contents;
    my $found_daemon_option = 0;

    open my $r_fh, "<$sendmail_mc" 
        or die "Cannot open the file $sendmail_mc: $!";
    while (my $line = <$r_fh>) {
        chomp ($line);
        push @sendmail_contents, $line;
        if ($line =~ /^DAEMON_OPTIONS/) {
            $found_daemon_option = 1;
        }
    }
    close $r_fh;

    my $fix_line = qq{DAEMON_OPTIONS(`Name=MTA-v6, Family=inet6')};
    if (! $found_daemon_option) {
        my $tmp_file = "$PREFIX/var/tmp/mta_config_fix.$$";
        open my $w_fh, ">$tmp_file" 
            or die "Cannot open the temp file $tmp_file: $!";

        my $found_filter_line = 0;
        for my $line (@sendmail_contents) {
            print $w_fh "$line\n";
            if ($line =~ /^INPUT_MAIL_FILTER/) {
                $found_filter_line = 1;                
                print $w_fh "$fix_line\n";
            }
        }
        print $w_fh "$fix_line\n" unless ($found_filter_line);
        close $w_fh;

        my ($mode, $uid, $gid) = (stat($sendmail_mc))[2,4,5];
        chmod S_IMODE($mode), $tmp_file;
        chown $uid, $gid, $tmp_file;

        if (! rename ($tmp_file, $sendmail_mc)) {
            unlink $tmp_file;
            die "Cannot rename $tmp_file to $sendmail_mc: $!";
        }
    } else {
        $self->log_warn(<<EOT);
Unable to convert your existing sendmail.mc file to use IPv6, please adjust
your DAEMON_OPTIONS directive to include the following settings:-

$fix_line
EOT
    }
    
    return 1;
}

sub mta_type {
    my ($self) = @_;

    if (!$self->{mta_type}) {
        if (-e "$PREFIX/sendmail") {
            $self->{mta_type} = 'sendmail';
        } elsif (-e "$PREFIX/postfix") {
            $self->{mta_type} = 'postfix';
        } else {
            $self->{mta_type} = 'notfound';
        }
    }
    return $self->{mta_type};
}

sub configlist_cmd {
    my ($self, $use_manifest) = @_;
    my $option = "";

    my $mta = $self->mta_type();
    if ($mta eq 'sendmail') {
        $option = "--sendmail";
    } elsif ($mta eq 'postfix') {
        $option = "--postfix";
    }

    my $manifest = $use_manifest ? $MANIFEST_FILE : '-';
    return "$SCRIPT_DIR/pmx-v5configlist $option $manifest";
}

sub process_files {
    my $self = shift;
    my %opts = @_;
    my $fh;

    $self->log_info1("Copying configuration files");
    $self->{MANIFEST} = {};

    my $prefix;
    my $fb_count = 1;
    my $copier = sub {
        my $line = shift || '';
        chomp($line);

        my ($tag, @params) = split "\t", $line;
        if ($tag =~ /^PREFIX:/) {
            $prefix = $params[0] || '';
            $prefix =~ s{/$}{};
            return;
        }
        die("Cannot relocate without prefix\n") unless $prefix;

        my ($uid, $gid, $mode, undef, $file) = @params;
        if ($uid ne '-') {
            if (!($uid =~ /^\d+$/ && $gid =~ /^\d+$/ && $mode =~ /^\d+$/ && $file)) {
                $self->log2file('WARNING', "Invalid file definition line: '$line'");
                return;
            }

            my $newfile = copy($file,
                uid      => $uid,
                gid      => $gid,
                mode     => $mode,
                relocate => $prefix,
            );
            if (defined $newfile) {
                $self->{MANIFEST}{$newfile} = 1;
                $self->log2file('INFO', "Transferred $newfile");
            }
            else {
                $self->log2file("INFO", "Not updating $PREFIX/$file: Not a regular file or directory");
            }
        }
        else {
            $self->log2file('INFO', "Not updating $PREFIX/$file");
        }
        $self->ping() unless ($fb_count++ % 100);
        return;
    };

    if ($opts{use_manifest}) {
        $self->run_as_pmxv5($self->configlist_cmd(1));
        open($fh, "<", $MANIFEST_FILE) or
            die("Could not open $MANIFEST_FILE: $!\n");
        $copier->($_) while (<$fh>);
        close($fh);
    }
    else {
        my %opts = ( filter => qr/^(MANIFEST|PREFIX):\t/ );
        $self->pipeline_as_pmxv5($self->configlist_cmd(0), $copier, %opts);
    }

    return 1;
}

sub finalize {
    my $self = shift;

    $self->log_info1("Finalizing changes");
    my $postfix_cf    = "$PREFIX/postfix/etc/main.cf";
    my $sendmail_cf   = "$PREFIX/sendmail/etc/mail/sendmail.mc";
    my $minisrv_cf    = "$PREFIX/etc/manager/manager.conf";
    my $pmx_cf        = "$PREFIX/etc/pmx.conf";
    my $internal_host = "$PREFIX/etc/internal-hosts";

    # MTA fixups
    if ($self->{MANIFEST}{$postfix_cf}) {
        $self->log_info2("Rebuilding postfix configs");
        $self->run_as_root("$PREFIX/postfix/sbin/pmx-postfix-conf --make");
    }
    elsif ($self->{MANIFEST}{$sendmail_cf}) {
        $self->log_info2("Rebuilding sendmail configs");
        $self->run_as_root("$PREFIX/sendmail/sbin/pmx-sendmail-conf --make");
    }

    if ($self->{MANIFEST}{$pmx_cf}) {
        my $mail_sender = `grep '^mail_sender' $pmx_cf`;
        chomp($mail_sender) if $mail_sender;
        if ($mail_sender) {
            # will replace
            $self->log_info2("Checking mail_sender value format");
            $self->run_as_pmxv6("$SCRIPT_DIR/pmx-v6configupdate");
        }
    }

    # Miniserv
    if ($self->{MANIFEST}{$minisrv_cf}) {
        unless (`grep '^ipv6=1' $minisrv_cf`) {
            $self->log2file('INFO', "Enabling IPv6 for minisrv");
            open(my $fh, ">>", $minisrv_cf) ||
                die("Cannot open: $minisrv_cf: $!\n");
            print {$fh} "ipv6=1\n";
            close($fh) ||
                die("Cannot close: $minisrv_cf: $!\n");
        }
    }

    # internal-hosts
    if ($self->{MANIFEST}{$internal_host}) {
        $self->log2file('INFO', "Adding ::1 to list of internal-hosts");
        $self->run_as_pmxv6("pmx-list add internal-hosts ::1");
    }

    $self->log_info2("Updating cache info");
    $self->run_as_pmxv6("pmx-config --validate-cache");

    $self->log_info2("Creating cdb files for maps, list and multilists");
    $self->run_as_pmxv6("pmx-makemap -a");

    # euwi skin fixup
    $self->euwi_skin_fix();

    if ($self->{has_db}) {
        my $dump_file  = "$PREFIX/var/tmp/config-tables-dump.sql";
        $self->log_info2('Migrating group data');
        $self->run_as_pmxv6("pmx-database stop",  exitcodes => [2]);
        $self->run_as_pmxv5("pmx-database start", exitcodes => [2]);
        $self->run_as_pmxv5("$SCRIPT_DIR/pmx-v5dumpgroupdata $dump_file");
        $self->run_as_pmxv5("pmx-database stop",  exitcodes => [2]);
        $self->log2file('INFO', "Group data has been dumped to $dump_file");
        $self->log2file('INFO', 'Loading group data to new database');
        $self->run_as_pmxv6("pmx-database start", exitcodes => [2]);
        $self->run_as_pmxv6("$PREFIX/postgres/bin/psql -f $dump_file pmx_quarantine");

        $self->log_info2('Updating resources');
        $self->run_as_pmxv6("$SCRIPT_DIR/pmx-v6resync to-db");
        $self->run_as_pmxv6("$SCRIPT_DIR/pmx-v6resync from-db");
    }
    else {
        $self->log_info2("No database detected on this instance of PureMessage. " .
                         "Skipping group data and database resources");
    }

    return 1;
}

sub validate {
    my $self = shift;

    if (-x "$PREFIX/bin/pmx-policy") {
        $self->log_info2("Checking policy configuration file");
        $self->run_as_pmxv6("pmx-policy check");
    }

    if (-e "$PREFIX/sendmail" || -e "$PREFIX/postfix") {
        $self->log_info2("Checking MTA");
        $self->run_as_pmxv6("pmx-service start smtp");
        $self->run_as_pmxv6("pmx-service status smtp | grep \"running => 1\"");
        $self->run_as_pmxv6("pmx-service stop smtp");
    }

    return 1;
}

sub euwi_skin_fix {
    my $self = shift;
    my $euwi_skin_dir = "$PREFIX/lib/manager/HTTPD/www/skins";
    my $transparent_image = "$IMAGE_DIR/transparent.gif";
    my $euwi_image = "end-user-web-quarantine.gif";

    return unless (-d $euwi_skin_dir);

    opendir (my $r_dir, $euwi_skin_dir) 
        or die "Cannot open directory $euwi_skin_dir: $!";
    while (my $entry = readdir $r_dir) {
        next if $entry =~ /^\.{1,2}$/;
        next if $entry =~ /^default$/;
       
        my $custom_skin_dir = "$euwi_skin_dir/$entry";
        if (-d $custom_skin_dir && ! -f "$custom_skin_dir/$euwi_image") {
            $self->log2file("INFO", "Found custom skin directory $entry. Copy a transparent image to it.");
            $self->run_as_pmxv6("cp $transparent_image $euwi_skin_dir/$entry/$euwi_image");
        }
    }
    closedir ($r_dir); 
}

1;
