package PureMessage::V6::Migrate::LogFile;
use strict;
use warnings;
use base 'PureMessage::V6::Migrate::Base';
use ActiveState::Prompt             qw(yes);
use PureMessage::Version            qw($PREFIX);
use PureMessage::V6::Migrate::Utils qw(copy get_free_space);
use File::Find qw(find);
use File::Path qw(rmtree);

our $MANIFEST_FILE = "$PREFIX/var/tmp/MIGRATION_LOG_MANIFEST.txt";
our $SCRIPT_DIR = "$PREFIX/var/migrate/v6";

sub display {
    my $self = shift;
    
    $self->show("Checking required diskspace...");

    my ($free, $required) = $self->check_required_space();
    if ($free < $required) {
        $self->show("Insufficient diskspace, required=$required " .
                    "available=$free");
        return 0;
    }    

    $self->show("Migrating your PureMessage 5 logs will overwrite your " .
                "current log search files and data.");

    if(yes("Would you like to proceed (yes/no)?", 1)) {
        return 1; 
    }
    return 0;
}

sub migrate {
    my $self = shift;
    my %opts = @_;

    if (!$self->process_files(%opts)) {
        $self->log_err("Copying files");
        return 0;
    }
    elsif (!$self->cleanup()) {
        $self->log_err("Cleanup unnecessary log files");
        return 0;
    }

    return 1;
}

#####

sub check_required_space {
    my $self = shift;
    my $fh;

    my $free = get_free_space("$PREFIX/var/log");
    die "Could not determine available space\n"
        unless (defined $free && $free =~ /^\d+$/);

    my $required;
    $self->pipeline_as_pmxv5("$SCRIPT_DIR/pmx-v5loglist --diskusage", 
        sub {
            chomp(my $line = shift || '');
            $required = (split /\t/, $line)[-1];
        },
        filter => qr/^LSIZE:\t/,
    );

    die "Could not determine required space\n"
        unless ($required && $required =~ /^\d+$/);
    
    return ($free, $required);
}

sub loglist_cmd {
    my ($self, $use_manifest) = @_;

    my $manifest = $use_manifest ? $MANIFEST_FILE : '-';
    return "$SCRIPT_DIR/pmx-v5loglist --manifest $manifest";
}

sub process_files {
    my $self = shift;
    my %opts = @_;
    my $fh;

    $self->log_info1("Copying log files from PureMessage 5 to PureMessage 6");
    
    my $fb_count = 1;
    my $copier = sub { 
        my $line = shift;
        chomp($line);

        my (undef, $uid, $gid, $mode, undef, $file) = split "\t", $line;
        if ($uid ne '-') {
            if (!($uid =~ /^\d+$/ && $gid =~ /^\d+$/ && $mode =~ /^\d+$/ && $file)) {
                $self->log2file('WARNING', "Invalid file definition line: '$line'");
                return;
            }

            my $newfile = copy($file,
                uid  => $uid,
                gid  => $gid,
                mode => $mode
            );
            if (defined $newfile) {
                push @{$self->{files}}, [$uid, $gid, $mode, undef, $file];
                $self->log2file("INFO", "Transferred $newfile");
            }
            else {
                $self->log2file("INFO", "Not updating $PREFIX/$file: Not a regular file or directory");
            }
        }
        else {
            $self->log2file("INFO", "Not updating $PREFIX/$file");
        }
        $self->ping() unless ($fb_count++ % 100);
    };

    if ($opts{use_manifest}) {
        $self->run_as_pmxv5($self->loglist_cmd(1));
        open $fh, '<', $MANIFEST_FILE or
            die("could not open $MANIFEST_FILE: $!\n");
        $copier->($_) while (<$fh>);
        close($fh);
    }
    else {
        my %opts = ( filter => qr/^MANIFEST:\t/ );
        $self->pipeline_as_pmxv5($self->loglist_cmd(0), $copier, %opts);
    }

    return 1;
}

sub cleanup {
    my $self = shift;

    $self->log_info1("Cleaning unnecessary log files");
    my $bases = ["$PREFIX/var/log", "$PREFIX/var/reports"];
    my %exclude  = ('var/log' => 1, 'var/reports' => 1);  
    my %manifest = map {$_->[4] => $_->[0]} @{$self->{files}};
    my @delete_dirs = ();

    my $wanted = sub {
        my $file = $File::Find::name;
        $file =~ s,^$PREFIX/,,;
        return if defined $manifest{$file};
        foreach my $cf (grep { $manifest{$_} ne '-' } sort keys %manifest) {
            next if defined $exclude{$cf};
            if ($file =~ qr/^$cf/) {
                if (-f "$PREFIX/$file") {
                    $self->log2file("INFO", "Removing file $PREFIX/$file");
                    rmtree("$PREFIX/$file");
                }
                elsif (-d "$PREFIX/$file") {
                    # Remove all directories at the end;
                    push @delete_dirs, "$PREFIX/$file";
                }
                last;
            }    
        }
    };

    my $preprocess = sub {
        return sort @_;
    };

    remove_files($bases, $wanted, $preprocess);
    foreach my $dir (@delete_dirs) {
        $self->log2file("INFO", "Removing directory $dir")
            if rmdir($dir);
    }

    return 1;
}

sub remove_files {
    my ($bases, $wanted, $preprocess) = @_;

    foreach my $base (@$bases) {
        find({
            wanted      => $wanted,
            preprocess  => $preprocess,
            no_chdir    => 1,
            follow_skip => 2,
        }, $base);
    }
}

1;
