package PureMessage::V6::Migrate::Quarantine::Dir;
use strict;
use warnings;
use base 'PureMessage::V6::Migrate::Base';
use File::Path                      qw(rmtree);
use PureMessage::Config             qw(%config);
use PureMessage::Version            qw($PREFIX);
use PureMessage::V6::Migrate::Utils qw(
    copy chuser_to_pmxv6 chuser_to_root
);

our $SCRIPT_DIR = "$PREFIX/var/migrate/v6";

sub migrate {
    my $self = shift;
    my %opts = @_;

    if (!$self->initialize_quarantine(%opts)) {
        $self->log_err("Unable to re-create quarantine");
        return 0;
    }
    elsif (!$self->process_files(%opts)) {
        $self->log_err("Copying files");
        return 0;
    }

    return 1;
}

sub initialize_quarantine {
    my $self = shift;

    my $quarantine_dir = $config{quarantine_dir};
    $self->log_info1("Initializing quarantine '$quarantine_dir'");

    eval {
        chuser_to_pmxv6();
        require PureMessage::Quarantine;

        rmtree($quarantine_dir);
        PureMessage::Quarantine->new("dir:$quarantine_dir",
            'index-type' => 'cdb',
             create => 1,
        );
    };
    chuser_to_root();

    if ($@) {
        $self->log_err($@);
        return 0;
    }
    return 1;
}

sub process_files {
    my $self = shift;
    my %opts = @_;
    my $fh;

    $self->log_info1("Copying quarantine files (This may take several hours to run)");

    my $feedback_chunks = 1000;
    if (my $qsize = $opts{QSIZE}) {
        my $num_files = int($qsize->{DIR} / 4096);
        $feedback_chunks = $num_files ? int($num_files / 100) : 1000;
        $feedback_chunks = 1000 if $feedback_chunks < 1000;
        $feedback_chunks -= $feedback_chunks % 1000;
    }

    $self->log_info2("Progress reported every $feedback_chunks files");

    my $file_count = 0;
    my $copier = sub {
        my $line = shift;
        chomp($line);

        my (undef, $uid, $gid, $mode, undef, $file) = split "\t", $line;

        my $newfile;
        if ($uid ne '-') {
            if (!($uid =~ /^\d+$/ && $gid =~ /^\d+$/ && $mode =~ /^\d+$/ && $file)) {
                $self->log2file('WARNING', "Invalid file definition line: '$line'");
                return;
            }

            $newfile = copy($file,
                uid => $uid,
                gid => $gid,
                mode => $mode
            );
        }
            
        if (defined $newfile) {
            $file_count++;

            $self->log_info2("Transferred $file_count file(s)")
                if ($file_count % $feedback_chunks) == 0;
        }
        else {
            my $err = $uid eq '-' ? 'Not required' : 'Not supported';
            $self->log2file('INFO', "Not updating $PREFIX/$file: $err");
        }
    };

    {
        my %opts = ( filter => qr/^MANIFEST:\t/ );
        $self->pipeline_as_pmxv5("$SCRIPT_DIR/pmx-v5quarantine", $copier, %opts);
    }

    $self->log_info2("Transferred $file_count file(s) in total");
    $self->log_info2("Quarantine migration complete");

    return 1;
}

1;
