package PureMessage::V6::Migrate::Utils;
use warnings;
use strict;
use base 'Exporter';
use Fcntl      qw(S_IMODE S_ISREG S_ISDIR O_BINARY O_RDONLY O_WRONLY O_CREAT O_TRUNC);
use PPM::PureMessage::Utils qw(pmx_user);
use ActiveState::Unix::DiskFree qw(df);
use POSIX qw(geteuid getegid);
use File::Find;

use constant BUFFERSIZE => 4096;

our @EXPORT_OK = qw(is_root copy cmd_runas_pmxv5 cmd_runas_pmxv6 cmd_runas_root
                    relocate_cmd is_pmxv5_installed get_free_space
                    chuser_to_pmxv5 chuser_to_pmxv6 chuser_to_root
                    chown_as_pmxv5 chown_as_pmxv6 get_pmxv6_user
                    get_pmxv5_user has_pmxv5_db has_pmxv6_db
                    );

our %GLOBALS;

my $CNT = 0;

sub is_root {
    $> == 0;
}

sub is_pmxv5_installed {
    my $globals = _get_globals();
    my $prefix = $globals->{pmx5_prefix};
    return 1 if (defined $prefix and -f "$prefix/etc/.pmx_installed");
    return 0;
}

sub has_pmxv5_db {
    my $globals = _get_globals();
    my $prefix = $globals->{pmx5_prefix};

    return 1 if (-f "$prefix/postgres/bin/psql");
    return 0;
}

sub has_pmxv6_db {
    my $globals = _get_globals();
    my $prefix = $globals->{pmx6_prefix};

    return 1 if (-f "$prefix/postgres/bin/psql");
    return 0;
}

sub copy {
    my ($from, %args) = @_;
    my $globals = _get_globals();

    my $uid = $args{uid} eq $globals->{pmx5_uid} ? $globals->{pmx6_uid} : $args{uid};
    my $gid = $args{gid} eq $globals->{pmx5_gid} ? $globals->{pmx6_gid} : $args{gid};
    my $relocate = $args{relocate} || '';

    my $pmx6_prefix = $globals->{pmx6_prefix};
    my $to = "$pmx6_prefix/$from";
    if (S_ISDIR($args{mode})) {
        _mkdir($to, $args{mode}, $uid, $gid);
    }
    elsif (S_ISREG($args{mode})) {
        my $tmp = sprintf "$to-$$.%010s", _cnt();
        my $pmx5_prefix = $globals->{pmx5_prefix};
        _copy("$pmx5_prefix/$from", $tmp, $args{mode}, $uid, $gid, $relocate);
        unless (rename($tmp, $to)) {
            my $err = "Can't rename: $tmp -> $to: $!\n";
            unlink $tmp or warn "Can't remove $tmp: $!\n";
            die $err;
        }
    }
    else {
        return;
    }

    return $to;
}

sub relocate_cmd {
    my $cmd = shift;
    my $globals = _get_globals();

    $cmd =~ s,%%PREFIX5%%,$globals->{pmx5_prefix},g;
    $cmd =~ s,%%PREFIX6%%,$globals->{pmx6_prefix},g;
    return $cmd;
}

sub cmd_runas_pmxv5 {
    my $cmd  = shift;
    my %opts = @_;
    my $globals = _get_globals();
    my $cmdline;

    $cmdline = sprintf("%s/bin/pmx-env ", $globals->{pmx5_prefix})
        unless $opts{raw_cmd};
    $cmdline .= relocate_cmd($cmd);

    if (wantarray) {
        return ('su', '-', $globals->{pmx5_user}, '-c', $cmdline);
    }
    else {
        return sprintf "su - %s -c '%s'", $globals->{pmx5_user}, $cmdline;
    }
}

sub cmd_runas_pmxv6 {
    my $cmd  = shift;
    my %opts = @_;
    my $globals = _get_globals();
    my $cmdline;

    $cmdline = sprintf("%s/bin/pmx-env ", $globals->{pmx6_prefix})
        unless $opts{raw_cmd};
    $cmdline .= relocate_cmd($cmd);

    if (wantarray) {
        return ('su', '-', $globals->{pmx6_user}, '-c', $cmdline);
    }
    else {
        return sprintf "su - %s -c '%s'", $globals->{pmx6_user}, $cmdline;
    }
}

sub cmd_runas_root  { return relocate_cmd(shift) }
sub chuser_to_pmxv5 { return _runas_user('pmx')  }
sub chuser_to_pmxv6 { return _runas_user('pmx6') }
sub chuser_to_root  { return _runas_user('root') }    

sub _runas_user {
    my $user = shift;

    my ($uid, $gid) = (0, 0);
    if ($user ne 'root') {
        my $globals = _get_globals();
        $uid = $user eq 'pmx' ? $globals->{pmx5_uid} : $globals->{pmx6_uid};
        $gid = $user eq 'pmx' ? $globals->{pmx5_gid} : $globals->{pmx6_gid};
    }

    $> = $uid;
    die "Unable to change to uid '$uid'\n"
        unless geteuid() == $uid;

    $) = $gid;
    die "Unable to change to gid '$gid'\n"
        unless getegid() == $gid;

    return 1;
}

sub get_free_space {
    my $dir = shift;
    my $info;
    eval {
        $info = df($dir);
    };
    if ($@) {
        die "Can't determine free disk space for directory $dir: $@";
    }
    return $info->{free};
}

sub chown_as_pmxv6 { _chown_as(shift, 'pmx6') };
sub chown_as_pmxv5 { _chown_as(shift, 'pmx') };

sub get_pmxv5_user {
    my $global = _get_globals();
    
    return $global->{pmx5_user};
}

sub get_pmxv6_user {
    my $global = _get_globals();
    
    return $global->{pmx6_user};
}

#
# Helper routines
#

sub _read_prefix_file {
    my $fh;
    my $opened;
    my $prefix;
    if (ref($_[0]) eq "GLOB") {
        $fh = $_[0];
    }
    else {
        open($fh, $_[0]) or die "Can't open $_[0]: $!";
        $opened++;
    }

    # use the last line in the file and remove trailing slash and newline 
    while(<$fh>) {
        chomp;
        $prefix = $_ if -d $_;
    }
    if ($prefix) {
        $prefix =~ s,/$,,;
    }

    if ($opened) {
        close $fh or warn "Couldn't close $_[0]: $!";
    }
    return $prefix;
}

sub _get_globals {
    return \%GLOBALS if %GLOBALS;
    my $pmx5_prefix = _get_prefix('puremessage-5');
    my $pmx6_prefix = _get_prefix('puremessage-6');
    my ($pmx5_user, $pmx5_uid, $pmx5_gid) = _get_owner($pmx5_prefix);
    my ($pmx6_user, $pmx6_uid, $pmx6_gid) = _get_owner($pmx6_prefix);

    $pmx5_prefix =~ s{/$}{} if defined $pmx5_prefix;
    $pmx6_prefix =~ s{/$}{} if defined $pmx6_prefix;
    %GLOBALS = (
        pmx5_prefix => $pmx5_prefix,
        pmx6_prefix => $pmx6_prefix,
        pmx5_user   => $pmx5_user,
        pmx5_uid    => $pmx5_uid,
        pmx5_gid    => $pmx5_gid,
        pmx6_user   => $pmx6_user,
        pmx6_uid    => $pmx6_uid,
        pmx6_gid    => $pmx6_gid,
    );

    return \%GLOBALS;
}

sub _get_prefix {
    my $prefix_file = shift;
    my $prefix;
    my $prefix_dir = defined $ENV{PMX_PREFIX_DIR} ? $ENV{PMX_PREFIX_DIR} : "/etc";
    $prefix_file = "$prefix_dir/$prefix_file";
    return unless defined $prefix_file and -e $prefix_file;

    open(my $fh, $prefix_file) or die "Can't open $prefix_file: $!";

    eval {
        $prefix = _read_prefix_file($fh);
    };
    if ($@) {
        die "Can't determince PureMessage install prefix: $@";
    }

    close $fh;
    die "The PureMessage install prefix cannot be found!\n" unless defined $prefix;
    die "No pmx user for prefix $prefix" unless -d "$prefix/home";
    return $prefix;
}

sub _get_owner {
    my $prefix = shift;
    my @pw = pmx_user($prefix);
    return @pw[0,2,3];
}

sub _copy {
    my ($from, $to, $perms, $uid, $gid, $v5_prefix) = @_;
    my $globals = _get_globals();
    $perms = S_IMODE($perms);

    sysopen (my $FROM, $from, O_BINARY|O_RDONLY) 
        or die "Can't read $from: $!\n";

    my $TO;
    my $orig_umask = umask(0);
    unless (sysopen ($TO, $to, O_BINARY|O_WRONLY|O_CREAT|O_TRUNC, $perms)) {
        my $err = $!;
        umask($orig_umask);
        die "Can't write $to: $err\n";
    }
    umask($orig_umask);
    if (chown($uid, $gid, $to) != 1) {
        die "Can't chown $to to $uid:$gid: $!\n";
    }

    my ($rlen, $wlen);
    my ($opfx, $npfx) = ($v5_prefix, $globals->{pmx6_prefix});
    my $tmpbuf = '';

    do {
       $rlen = sysread($FROM, my $inbuf, BUFFERSIZE);
       die "Could not read '$from': $!\n" unless defined $rlen;

       my $outbuf = $tmpbuf . $inbuf;

       # Need to relocate?
       if ($opfx) {
           $outbuf =~ s{$opfx/}{$npfx/}g;
           $wlen   = length $outbuf;
           if ($rlen == BUFFERSIZE) {
               my $pref_len = length $opfx;
               $wlen -= $pref_len if $wlen > $pref_len;
           }
       }
       else {
           $wlen = length $outbuf;
       }

       my $rc = syswrite($TO, $outbuf, $wlen);
       die "Could not write '$to': $!\n" unless defined $rc;

       $tmpbuf = substr($outbuf, $rc);
    } until ($rlen != BUFFERSIZE);

    close $FROM or die "Can't close $from: $!\n";
    close $TO   or die "Can't close $to: $!\n";

    return;
}

sub _mkdir {
    my ($dir, $perms, $uid, $gid) = @_;
    $perms = S_IMODE($perms);

    my $omask = umask(0);
    if (!mkdir($dir, $perms)) {
        unless ($!{EEXIST}) {
            umask($omask);
            die "mkdir '$dir' failed: $!\n";
        }
    }
    else {
        chown($uid, $gid, $dir) or die "Can't chown $dir: $!";
    }
    umask($omask);
    return $dir;
}

sub _cnt {
    $CNT = -1 if $CNT >= 999999;
    $CNT++;
}

sub _chown_as {
    my ($dir, $user) = @_;
    $dir = relocate_cmd($dir);

    my $globals = _get_globals();

    my $uid = $user eq 'pmx' ? $globals->{pmx5_uid} : $globals->{pmx6_uid};
    my $gid = $user eq 'pmx' ? $globals->{pmx5_gid} : $globals->{pmx6_gid};

    my @all_files;
    my $wanted = sub {
        push @all_files, $File::Find::name;
    };
    find({
        wanted      => $wanted,
        no_chdir    => 1,
        follow_fast => 1,
    }, $dir);

    chown $uid, $gid, @all_files;
}

1;

